#!/usr/bin/perl -w
use English; 		# things like MATCH
use Getopt::Long;
use File::Basename;	# basename, dirname
use Fcntl;		# O_RDWR

# A Perl script to SYS the boot sector of a disk image or
# a device for FreeDOS. Do not forget to copy kernel.sys
# and command.com to the root directory of the target!
# For the latter, you can use mtools, or just mount the
# target. If needed, use loopback mounting. Sample line
# for ~/.mtoolsrc: drive m: file="/tmp/test.img"

# Perl by Eric Auer, 16aug2004. The boot sectors are by
# the FreeDOS project and licensed under the GNU GPL.
# This Perl script itself is free Public Domain software.

# TODO: hidden sectors / heads / sec_per_track / OEM ID
# override should be possible with command line options.
# TODO: allow some user-specified seek to skip over any
# headers / partition table, e.g for DOSEMU disk images.

my ($disk, $lba, $help);	# options
my ($fsboot, $bootcode);	# strings used as arrays
my ($nasmopt, $bootsource);	# strings
my ($i, $fat);			# numbers

my $bootdir = " " . dirname($0) . "/bootsecs/";	# boot sector source dir

my $result = GetOptions("disk:s"=>\$disk, "lba!"=>\$lba, "help!"=>\$help);
# "o"=>\$oflag, "verbose"=>\$verbosebool, "string=s"=>\$stringmandatory
# "string:s"=>stringoptional ... similar for i integer and f float.
# @ARGV contains the unprocessed rest if $ARGV[0] after this.

if (($help) || (!defined $disk)) {
    print "FreeDOS boot sector SYS for Linux, v0.1, public domain, 2004.\n";
    print "  Puts a FreeDOS boot sector on a FAT (12/16/32) filesystem.\n";
    print "  You have to copy the kernel.sys and command.com file, too.\n";
    print "Usage: " . basename($0) . " --disk=file [--lba]\n";
    print "Options:\n  --disk=file_or_device target filesystem\n";
    print "  --lba selects LBA boot sectors\n";
    exit(1);
}

sysopen(IMAGE, $disk, O_RDWR)
    || die "cannot open filesystem $disk for read/write access";
binmode(IMAGE);		# not normally needed
if (sysread(IMAGE, $fsboot, 512) != 512) {
    die "cannot read boot sector from filesystem $disk";
}
sysseek(IMAGE, 0, 0) || die "cannot rewind to filesystem start";

if (substr($fsboot, 510, 2) ne "\x55\xaa") {
    die "boot sector magic value missing in filesystem";
}

if (substr($fsboot, 0x0b, 2) ne "\x00\x02") {
    die "not 512 bytes per sector\n";
}

if (substr($fsboot, 0x11, 2) eq "\x00\x00") {
	# other possible detection: 16 bit FAT1x size at 0x16 is zero
	$fat = 32;	# FAT32: zero FAT1x root directory entries
	if (($lba) && (substr($fsboot, 0x1c) eq "\x00\x00\x00\x00")) {
	    print STDERR "FAT32 LBA warning: Hidden sector count is zero!\n";
	}
} else {
	$fat = 16;
	# if 16 bit FAT1x size at 0x16 is at least 16, we have FAT16
	# actually it would be more correct to use the sector count
	# (possibly the 32 bit one), subtract reserved (boot) sector
	# count and fat count * fat size, and divide by sec per clust
	# and finally check if the result is at least 4096.

        my $fatsize = unpack('v',substr($fsboot,0x16,2));
	# print STDERR "FAT1x size: $fatsize\n";
	# letters: c/C signed/unsigned 8 bit, s/S 16, l/L 32...
	# network / big endian: n/N (16/32), VAX / little endian: v/V.
	if ($fatsize < 16) { $fat = 12; }	# FAT very small: FAT12
}

$bootsource = $bootdir . "boot.asm";	# source for FAT12 and FAT16
if ($fat == 12) { $nasmopt = "-dISFAT12"; }	# can do both CHS and LBA
if ($fat == 16) { $nasmopt = "-dISFAT16"; }	# can do both CHS and LBA
if ($fat == 32) {
    $nasmopt = "";				# no options for FAT32
    $bootsource = $bootdir . "boot32.asm";	# CHS version
    if ($lba) { $bootsource =~ s/32/32lb/; }	# LBA version
}

print STDERR "DOS boot sector for $disk will be created by:\n";
print STDERR "\tnasm -o /dev/stdout $nasmopt $bootsource\n";
open(BOOTSECT, "nasm -o /dev/stdout $nasmopt $bootsource |")
    || die "cannot fork";
binmode(BOOTSECT);	# not normally needed
if (sysread(BOOTSECT, $bootcode, 512) != 512) {
    die "boot sector compilation problem";
}
close(BOOTSECT) || die "boot sector $bootsource nasm error $?";

if (substr($bootcode, 510, 2) ne "\x55\xaa") {
    die "magic value missing in compiled boot sector";
}

substr($bootcode, 3, 8, "LINUX4.1");		# place OEM ID, the
						# "4.1" is to please M$
if ($fat != 32) {
    $i = 0x3e-11;				# size of classic BPB
} else {
    $i = 0x5a-11;				# size of FAT32 xBPB
}
substr($bootcode, 3+8, $i, substr($fsboot, 3+8, $i));	# copy BPB

print STDERR "Using FAT$fat, " . (($lba) ? "LBA, " : "") . "boot code"
    . " starts at offset " . sprintf("%x", $i+11) . ". Extended info:"
    . sprintf(" drive %x,\n", unpack('c', substr($fsboot, 11+$i-26, 1)))
    . "res. ", unpack('c', substr($fsboot, 11+$i-25, 1)) . ", "
    . sprintf("sig. %x, ", unpack('c', substr($fsboot, 11+$i-24, 1)))
    . sprintf("serno %X-%X, ",
	unpack('v', substr($fsboot, 11+$i-21, 2)),
	unpack('v', substr($fsboot, 11+$i-23, 2)))
    . "name [" . substr($fsboot, 11+$i-19, 11) . "], "	# label, if not at \
    . "type [" . substr($fsboot,11+$i-8,8) . "].\n";

syswrite(IMAGE, $bootcode, 512) || die "could not write updated boot sector";
close(IMAGE);

print STDERR "Boot sector successfully updated.\n";
exit(0);
